//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Item/BeamWavelengthItem.cpp
//! @brief     Implements class BeamWavelengthItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Item/BeamWavelengthItem.h"
#include "GUI/Model/MakeItem/DistributionItemCatalog.h"

namespace {

const double default_wl = 0.1;

void initDistribution(DistributionItem* newDistribution,
                      const DistributionItem* currentDistribution)
{
    newDistribution->setLimits(RealLimits::positive());
    newDistribution->setUnit(Unit::nanometer);

    double meanVal = default_wl;
    if (auto* cd = dynamic_cast<const SymmetricResolutionItem*>(currentDistribution))
        meanVal = cd->mean();

    if (auto* nd = dynamic_cast<SymmetricResolutionItem*>(newDistribution)) {
        nd->setMean(meanVal);
        nd->setMeanDecimals(4);
    }

    newDistribution->initDistribution(meanVal);
};

} // namespace

BeamWavelengthItem::BeamWavelengthItem()
    : BeamWavelengthItem(false)
{
}

BeamWavelengthItem::BeamWavelengthItem(bool onlySymmetricDistributions)
{
    m_distribution.init<DistributionItemCatalog>("Distribution", "", "distribution",
                                                 onlySymmetricDistributions
                                                     ? DistributionItemCatalog::symmetricTypes()
                                                     : DistributionItemCatalog::types(),
                                                 initDistribution);
}

double BeamWavelengthItem::wavelength() const
{
    return BeamDistributionItem::meanValue();
}

// ------------------------------------------------------------------------------------------------

SpecularBeamWavelengthItem::SpecularBeamWavelengthItem()
    : BeamWavelengthItem(true)
{
}

void SpecularBeamWavelengthItem::setToRange(const RealLimits& limits)
{
    if (auto* symmetricDistribution = dynamic_cast<SymmetricResolutionItem*>(distribution())) {
        if (!limits.isInRange(wavelength())) {
            const double new_value =
                limits.isLimited() ? (limits.upperLimit() - limits.lowerLimit()) / 2. : default_wl;
            symmetricDistribution->setMean(new_value);
        }
        symmetricDistribution->setLimits(limits);
    }
}
