//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Item/PointwiseAxisItem.cpp
//! @brief     Implements pointwise axis item
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Item/PointwiseAxisItem.h"
#include "Base/Axis/PointwiseAxis.h"
#include "Device/Coord/ICoordSystem.h"
#include "Device/Data/OutputData.h"
#include "Device/Histo/IntensityDataIOFactory.h"
#include "Device/InputOutput/OutputDataReadWriteINT.h"
#include "GUI/Model/Item/InstrumentItems.h"

PointwiseAxisItem::PointwiseAxisItem()
    : BasicAxisItem(M_TYPE)
    , m_instrument(nullptr)
{
    lowerBoundItem()->setEnabled(false);
    binsItem()->setEnabled(false);
    upperBoundItem()->setEnabled(false);
    addProperty(P_NATIVE_AXIS_UNITS, "nbins");
}

PointwiseAxisItem::~PointwiseAxisItem() = default;

void PointwiseAxisItem::init(const IAxis& axis, const QString& units_label)
{
    m_axis = std::unique_ptr<IAxis>(axis.clone());
    setItemValue(P_NATIVE_AXIS_UNITS, units_label);
    findInstrument();
}

const IAxis* PointwiseAxisItem::axis() const
{
    return m_axis.get();
}

QString PointwiseAxisItem::getUnitsLabel() const
{
    return getItemValue(P_NATIVE_AXIS_UNITS).toString();
}

std::unique_ptr<IAxis> PointwiseAxisItem::createAxis(double scale) const
{
    if (!checkValidity())
        return nullptr;

    auto* const converter = m_instrument->createCoordSystem();
    const auto converted_axis = converter->createConvertedAxis(0, Axes::Coords::DEGREES);

    // applying scaling
    std::vector<double> centers = converted_axis->binCenters();
    std::for_each(centers.begin(), centers.end(), [scale](double& value) { value *= scale; });

    return std::make_unique<PointwiseAxis>(converted_axis->axisName(), std::move(centers));
}

QByteArray PointwiseAxisItem::serializeBinaryData() const
{
    if (!m_axis)
        return {};

    OutputData<double> axisData;
    axisData.addAxis(*m_axis);

    std::stringstream ss;
    OutputDataReadWriteINT().writeOutputData(axisData, ss);
    return QByteArray(ss.str().c_str(), static_cast<int>(ss.str().size()));
}

void PointwiseAxisItem::deserializeBinaryData(const QByteArray& data)
{
    if (data.isEmpty())
        return;

    std::istringstream str(data.toStdString());
    std::unique_ptr<OutputData<double>> d(OutputDataReadWriteINT().readOutputData(str));
    m_axis = std::unique_ptr<IAxis>(d->axis(0).clone());
    findInstrument();
}

bool PointwiseAxisItem::checkValidity() const
{
    return m_axis && m_instrument && getUnitsLabel() != "nbins";
}

void PointwiseAxisItem::findInstrument()
{
    SessionItem* parent_item = parent();
    while (parent_item && !dynamic_cast<SpecularInstrumentItem*>(parent_item))
        parent_item = parent_item->parent();
    m_instrument = dynamic_cast<SpecularInstrumentItem*>(parent_item);
}

void PointwiseAxisItem::updateIndicators()
{
    if (!checkValidity())
        return;

    auto* const converter = m_instrument->createCoordSystem();
    setLowerBound(converter->calculateMin(0, Axes::Coords::DEGREES));
    setUpperBound(converter->calculateMax(0, Axes::Coords::DEGREES));
    setBinCount(static_cast<int>(m_axis->size()));

    emitDataChanged();
}
