//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Item/SpecularBeamInclinationItem.cpp
//! @brief     Implements class SpecularBeamInclinationItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Item/SpecularBeamInclinationItem.h"
#include "Base/Axis/IAxis.h"
#include "Base/Const/Units.h"
#include "GUI/Model/Item/PointwiseAxisItem.h"
#include "GUI/Model/MakeItem/DistributionItemCatalog.h"

namespace {

void setAxisPresentationDefaults(BasicAxisItem* axisItem)
{
    ASSERT(axisItem);

    axisItem->setTitle("alpha_i");
    axisItem->binsItem()->setToolTip("Number of points in scan");
    axisItem->lowerBoundItem()->setToolTip("Starting value [deg]");
    axisItem->upperBoundItem()->setToolTip("Ending value [deg]");
    axisItem->lowerBoundItem()->setLimits(RealLimits::limited(0., 90.));
    axisItem->upperBoundItem()->setLimits(RealLimits::limited(0., 90.));

    if (!dynamic_cast<PointwiseAxisItem*>(axisItem)) {
        axisItem->setLowerBound(0.0);
        axisItem->setUpperBound(3.0);
        axisItem->setBinCount(500);
    }
}

void initDistribution(DistributionItem* newDistribution,
                      const DistributionItem* currentDistribution)
{
    newDistribution->setLimits(RealLimits::limited(-90.0, 90.0));
    newDistribution->setUnit(Unit::degree);

    double meanVal = 0.0;
    if (auto* cd = dynamic_cast<const SymmetricResolutionItem*>(currentDistribution))
        meanVal = cd->mean();

    if (auto* nd = dynamic_cast<SymmetricResolutionItem*>(newDistribution)) {
        nd->setMean(meanVal);
        nd->setMeanDecimals(3);
    }

    newDistribution->initDistribution(meanVal);
};


} // namespace

SpecularBeamInclinationItem::SpecularBeamInclinationItem()
    : BeamDistributionItem()
{
    m_distribution.init<DistributionItemCatalog>("Distribution", "", "distribution",
                                                 DistributionItemCatalog::symmetricTypes(),
                                                 initDistribution);

    m_uniformAlphaAxis.reset(new BasicAxisItem());
    setAxisPresentationDefaults(m_uniformAlphaAxis.get());
    m_currentAxisIsUniformAxis = true;
}

double SpecularBeamInclinationItem::scaleFactor() const
{
    return Units::deg;
}

BasicAxisItem* SpecularBeamInclinationItem::alphaAxis() const
{
    return m_currentAxisIsUniformAxis ? m_uniformAlphaAxis.get() : m_pointwiseAlphaAxis.get();
}

bool SpecularBeamInclinationItem::pointwiseAlphaAxisDefined() const
{
    return m_pointwiseAlphaAxis.get() != nullptr;
}

void SpecularBeamInclinationItem::selectUniformAxis()
{
    m_currentAxisIsUniformAxis = true;
}

void SpecularBeamInclinationItem::selectPointwiseAxis()
{
    ASSERT(pointwiseAlphaAxisDefined());
    m_currentAxisIsUniformAxis = false;
}

void SpecularBeamInclinationItem::initUniformAxis(const IAxis& axis)
{
    m_uniformAlphaAxis->setBinCount(static_cast<int>(axis.size()));
}

void SpecularBeamInclinationItem::initPointwiseAxis(const IAxis& axis, QString units)
{
    if (!m_pointwiseAlphaAxis) {
        m_pointwiseAlphaAxis.reset(new PointwiseAxisItem());
        setAxisPresentationDefaults(m_pointwiseAlphaAxis.get());
    }

    m_pointwiseAlphaAxis->init(axis, units);
    m_pointwiseAlphaAxis->updateIndicators();
}
