//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Model/InstrumentModel.cpp
//! @brief     Implements class InstrumentModel
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Model/InstrumentModel.h"
#include "GUI/Model/Item/InstrumentItems.h"
#include "GUI/Model/MakeItem/InstrumentItemCatalog.h"
#include "GUI/Model/Model/SessionXML.h"
#include "GUI/Model/XML/Serializer.h"
#include "GUI/Util/String.h"
#include <QUuid>

InstrumentModel::~InstrumentModel()
{
    qDeleteAll(m_instruments);
}

void InstrumentModel::clear()
{
    qDeleteAll(m_instruments);
    m_instruments.clear();
}

void InstrumentModel::serialize(Serializer& s)
{
    s.assertVersion(0);
    s.rw<InstrumentItemCatalog>("InstrumentItems", m_instruments);
}

InstrumentItem* InstrumentModel::insertCopy(const InstrumentItem& source)
{
    auto* copy = source.createCopy();
    copy->setId(QUuid::createUuid().toString());
    m_instruments << copy;
    return copy;
}

QVector<InstrumentItem*> InstrumentModel::instrumentItems() const
{
    return m_instruments;
}

QStringList InstrumentModel::instrumentNames() const
{
    QStringList existingNames;
    for (const auto* item : instrumentItems())
        existingNames << item->instrumentName();
    return existingNames;
}

QString InstrumentModel::suggestInstrumentName(const QString& baseName) const
{
    return GUI::Util::String::suggestName(instrumentNames(), baseName);
}

QVector<InstrumentItem*>
InstrumentModel::instrumentItems(const std::function<bool(const InstrumentItem*)>& accept) const
{
    QVector<InstrumentItem*> result;
    for (auto* p : instrumentItems())
        if (accept(p))
            result << p;

    return result;
}

QVector<Instrument2DItem*> InstrumentModel::instrument2DItems() const
{
    QVector<Instrument2DItem*> result;
    for (auto* p : instrumentItems())
        if (auto* p2D = dynamic_cast<Instrument2DItem*>(p))
            result << p2D;

    return result;
}

InstrumentItem* InstrumentModel::findInstrumentById(const QString& instrumentId) const
{
    for (auto* instrument : instrumentItems())
        if (instrument->id() == instrumentId)
            return instrument;

    return nullptr;
}

bool InstrumentModel::instrumentExists(const QString& instrumentId) const
{
    return findInstrumentById(instrumentId) != nullptr;
}

void InstrumentModel::removeInstrument(InstrumentItem* instrument)
{
    m_instruments.removeAll(instrument);
    delete instrument;
}
