//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/DistributionEditor.cpp
//! @brief     Implements classes DistributionSelector and DistributionEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/DistributionEditor.h"
#include "GUI/Model/Item/BeamDistributionItem.h"
#include "GUI/Model/Item/DistributionItems.h"
#include "GUI/Model/MakeItem/DistributionItemCatalog.h"
#include "GUI/Model/Model/SessionModel.h"
#include "GUI/Model/Types/DoubleDescriptor.h"
#include "GUI/View/Instrument/DistributionForms.h"
#include "GUI/View/Instrument/EditDistributionDialog.h"
#include "GUI/View/PropertyEditor/GroupInfoBox.h"
#include "GUI/View/Tool/WidgetUtils.h"
#include <QComboBox>
#include <QFormLayout>
#include <QStackedLayout>
#include <utility>

namespace {

// some temporary functions to be able to assign the value of a distribution to an other

void assignBasic(const DistributionItem* src, DistributionItem* dst)
{
    dst->setNumberOfSamples(src->numberOfSamples());
    if (src->hasSigmaFactor())
        dst->setSigmaFactor(src->sigmaFactor());
}

void assignSymmetric(const SymmetricResolutionItem* src, SymmetricResolutionItem* dst,
                     const bool assignMean)
{
    assignBasic(src, dst);
    if (assignMean)
        dst->setMean(src->mean());
}

void assign(const DistributionNoneItem* src, DistributionNoneItem* dst, const bool assignMean)
{
    if (assignMean)
        dst->setMean(src->mean());
}

void assign(const DistributionGateItem* src, DistributionGateItem* dst)
{
    assignBasic(src, dst);
    dst->setRange(src->minimum(), src->maximum());
}

void assign(const DistributionLorentzItem* src, DistributionLorentzItem* dst, const bool assignMean)
{
    assignSymmetric(src, dst, assignMean);
    dst->setHwhm(src->hwhm());
}

void assign(const DistributionGaussianItem* src, DistributionGaussianItem* dst,
            const bool assignMean)
{
    assignSymmetric(src, dst, assignMean);
    dst->setStandardDeviation(src->standardDeviation());
}

void assign(const DistributionLogNormalItem* src, DistributionLogNormalItem* dst)
{
    assignBasic(src, dst);
    dst->setMedian(src->median());
    dst->setScaleParameter(src->scaleParameter());
}

void assign(const DistributionCosineItem* src, DistributionCosineItem* dst, const bool assignMean)
{
    assignSymmetric(src, dst, assignMean);
    dst->setSigma(src->sigma());
}

void assign(const DistributionTrapezoidItem* src, DistributionTrapezoidItem* dst)
{
    assignBasic(src, dst);
    dst->setCenter(src->center());
    dst->setLeftWidth(src->leftWidth());
    dst->setMiddleWidth(src->middleWidth());
    dst->setRightWidth(src->rightWidth());
}

void assign(const BeamDistributionItem* src, BeamDistributionItem* dst, const bool assignMean)
{
    const auto type = DistributionItemCatalog::type(src->distribution());
    auto* newDistribution = DistributionItemCatalog::create(type);
    dst->setDistribution(newDistribution);
    if (src->distribution()->is<DistributionNoneItem>())
        assign(dynamic_cast<const DistributionNoneItem*>(src->distribution()),
               dynamic_cast<DistributionNoneItem*>(dst->distribution()), assignMean);
    else if (src->distribution()->is<DistributionGateItem>())
        assign(dynamic_cast<const DistributionGateItem*>(src->distribution()),
               dynamic_cast<DistributionGateItem*>(dst->distribution()));
    else if (src->distribution()->is<DistributionCosineItem>())
        assign(dynamic_cast<const DistributionCosineItem*>(src->distribution()),
               dynamic_cast<DistributionCosineItem*>(dst->distribution()), assignMean);
    else if (src->distribution()->is<DistributionLorentzItem>())
        assign(dynamic_cast<const DistributionLorentzItem*>(src->distribution()),
               dynamic_cast<DistributionLorentzItem*>(dst->distribution()), assignMean);
    else if (src->distribution()->is<DistributionGaussianItem>())
        assign(dynamic_cast<const DistributionGaussianItem*>(src->distribution()),
               dynamic_cast<DistributionGaussianItem*>(dst->distribution()), assignMean);
    else if (src->distribution()->is<DistributionLogNormalItem>())
        assign(dynamic_cast<const DistributionLogNormalItem*>(src->distribution()),
               dynamic_cast<DistributionLogNormalItem*>(dst->distribution()));
    else if (src->distribution()->is<DistributionTrapezoidItem>())
        assign(dynamic_cast<const DistributionTrapezoidItem*>(src->distribution()),
               dynamic_cast<DistributionTrapezoidItem*>(dst->distribution()));
    else
        ASSERT(false);
}

} // namespace

//==================================================================================================
// DistributionSelector
//==================================================================================================

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

DistributionSelector::DistributionSelector(std::optional<MeanConfig> mean_config,
                                           GUI::ID::Distributions distributions, QWidget* parent,
                                           BeamDistributionItem* item)
    : QWidget(parent)
    , m_item(item)
    , m_meanConfig(std::move(mean_config))
    , m_distributions(distributions)
{
    ASSERT(item);
    m_formLayout = new QFormLayout(this);
    m_formLayout->setMargin(0);

    auto* distributionCombo =
        GUI::Util::createSelectionCombo(this, item->distributionSelection(), [=](int) {
            createDistributionWidgets();
            emit distributionChanged();
        });
    m_formLayout->addRow("Distribution:", distributionCombo);

    createDistributionWidgets();
}

void DistributionSelector::createDistributionWidgets()
{
    while (m_formLayout->rowCount() > 1)
        m_formLayout->removeRow(1);

    DistributionForm* form = nullptr;
    if (m_item->distribution()->is<DistributionCosineItem>())
        form = new CosineDistributionForm(m_meanConfig, this);
    else if (m_item->distribution()->is<DistributionGateItem>())
        form = new GateDistributionForm(this);
    else if (m_item->distribution()->is<DistributionGaussianItem>())
        form = new GaussDistributionForm(m_meanConfig, this);
    else if (m_item->distribution()->is<DistributionLogNormalItem>())
        form = new LogNormalDistributionForm(this);
    else if (m_item->distribution()->is<DistributionLorentzItem>())
        form = new LorentzDistributionForm(m_meanConfig, this);
    else if (m_item->distribution()->is<DistributionNoneItem>())
        form = new NoneDistributionForm(m_meanConfig, this);
    else if (m_item->distribution()->is<DistributionTrapezoidItem>())
        form = new TrapezoidDistributionForm(this);

    ASSERT(form);
    form->setupDistribution(m_item);
    m_formLayout->addRow(form);
    connect(form, &CosineDistributionForm::distributionChanged, this,
            &DistributionSelector::distributionChanged);
}

BeamDistributionItem* DistributionSelector::item() const
{
    return m_item;
}

const std::optional<MeanConfig>& DistributionSelector::meanConfig() const
{
    return m_meanConfig;
}

GUI::ID::Distributions DistributionSelector::distributions() const
{
    return m_distributions;
}

void DistributionSelector::refresh()
{
    createDistributionWidgets();
}


//==================================================================================================
// DistributionEditor
//==================================================================================================

DistributionEditor::DistributionEditor(const QString& title,
                                       const std::optional<MeanConfig>& mean_config,
                                       GUI::ID::Distributions distributions, QWidget* parent,
                                       BeamDistributionItem* item)
    : QWidget(parent)
    , m_title(title)
{
    auto* layout = new QVBoxLayout(this);

    auto* gbox = new GroupInfoBox(title, this);
    auto* glayout = new QVBoxLayout(gbox);
    m_selector = new DistributionSelector(mean_config, distributions, gbox, item);
    glayout->addWidget(m_selector);
    layout->addWidget(gbox);

    connect(m_selector, &DistributionSelector::distributionChanged, this,
            &DistributionEditor::distributionChanged);
    connect(gbox, &GroupInfoBox::clicked, this, &DistributionEditor::showDialog);
}

void DistributionEditor::updateData()
{
    m_selector->refresh();
}

void DistributionEditor::showDialog()
{
    showEditDistributionDialog(m_selector, m_title);
}

//--------------------------------------------------------------------------------------------------
// free functions
//--------------------------------------------------------------------------------------------------

void showEditDistributionDialog(DistributionSelector* selector, const QString& title)
{
    BeamDistributionItem* item = selector->item();
    if (!item)
        return;

    bool modifiedInDialog = false;

    SessionModel dummy("dummy");
    BeamDistributionItem* distr_item =
        dynamic_cast<BeamDistributionItem*>(dummy.copy(dynamic_cast<SessionItem*>(item)));
    EditDistributionDialog dialog(distr_item, title, selector->meanConfig(),
                                  selector->distributions(), selector);

    QObject::connect(&dialog, &EditDistributionDialog::distributionChanged,
                     [&] { modifiedInDialog = true; });
    if (dialog.exec() == QDialog::Accepted && modifiedInDialog) {
        assign(distr_item, item, selector->meanConfig().has_value());
        selector->refresh();
        selector->distributionChanged();
    }
}
